#!//usr/local/bin/munger

; Munger Copyright (c) 2004, 2005, James Bailie.
; All rights reserved.
; 
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are met:
; 
;     * Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following disclaimer.
;     * Redistributions in binary form must reproduce the above copyright
; notice, this list of conditions and the following disclaimer in the
; documentation and/or other materials provided with the distribution.
;     * The name of James Bailie may not be used to endorse or promote 
; products derived from this software without specific prior written permission.
; 
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
; LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.

; **************************************************************************
;
; To use this script with Apache, the FastCGI interface must have been
; compiled into your munger executable, and the FastCGI Apache module must
; have been installed and properly configured, first.
; 
; *************************************************************************

; This FastCGI script can perform expansion of embedded Munger code in
; static files, allowing the user to create dynamic web pages by embedding
; the code to produce the dynamic parts of a page inside the static
; elements of the page.  To use this script, two lines must be added to
; httpd.conf to associate a filename suffix with this script.  The relevant
; portions of my httpd.conf are listed below as an example.

; AddType application/x-munger-filtered .filtered
; Action application/x-munger-filtered /cgi-bin/expander.fcgi

; In addition, the location of the this script (expander.fcgi) must have
; been defined as a place from which CGI programs can be run with the
; ScriptAlias directive.

; Any file in the DocumentRoot directory with the .filtered suffix will
; cause this script to be run with the PATH_TRANSLATED environment variable
; set to the original filename.

# Embedded Munger Processor

# This filter outputs all lines it reads excepting lines beginning with a
# semicolon in column zero, an octothorpe in column zero, and ranges of
# contiguous lines beginning with a line containing an opening parenthesis in
# column zero, and ending with a line containing a semicolon in column zero.

# Lines beginning with a semicolon are considered comments and discarded.
# Lines beginning with an octothorpe are "include" lines, specifying a file
# whose content is to be inserted into the output stream.  Include lines
# must be of the form:

# #include "/path/to/filename"

# Blocks of lines beginning with an open parenthesis and terminated with a
# comment line, are sent to an inferior munger process for evaluation.  If
# a block was terminated by a comment line containg two contiguous
# semi-colons, then the interpreter will block waiting for a response from
# the inferior munger, and the response will then be inserted into the
# document being processed, replacing the code block and the terminating
# comment line.  To prevent return values from being mixed-up with
# side-effects, the "noprinter" intrinsic is invoked at startup on the
# inferior munger.  Blocks of code should refrain from emitting output
# until all the output can be sent at once, at the end of the block.
# Larger blocks can be broken up into smaller blocks at points where output
# is generated, to allow intermittent output to be interspersed with other
# processing.  If there is a one second or greater pause in the response 
# stream from the inferior munger, this script will assume the inferior 
# Munger has finished producing output for the expression sent to it and 
# move on to processing succeeding lines of the input file.

# If an expression produces no side-effects, no output will be produced
# from the inferior munger when the expression is evaluated.  To embed
# these expressions in a document, their blocks MUST be terminated with one
# semi-colon only, or this script will block forever trying to read a
# response from the inferior munger.

# There is no means whereby the filter can be made to escape or pass-through
# lines starting with (, #, or ;.  To cause the filter to output lines
# starting with these characters, one must write code to do so,  like this:  

# (print "; comment")
# (newline)
# ;;

# Although this script is a FastCGI script and must therefore use the
# "cgi_*" function to perform I/O, the template file fed to this script
# MUST NOT use these functions, but the ordinary I/O functions.

# The terminator(s) of a line of literal text may be suppressed by appending a
# backslash to the line.  This can be used to inline lisp output:  

# Version \
# (letn ((v (version))
#        (maj (car v))
#        (min (cadr v)))
#  (print maj "." min))
# ;;
# of \
# (print (date))
# ;;
# .

# Will result in the following output:

# Version 3.67 of Tue, 15 Jun 2004 13:33:25 EDT.

# To output a line with a trailing backslash, use code:

# (print "line of text\\")
# (newline)
# ;;

# **************************************************************************
#
# The following environment variables are propagated to the inferior
# Munger, and may therefore be accessed by embedded code in the template
# file.  Note that if a variable is unset, this will be indicated by
# setting it to the empty string in the inferior Munger.

# HTTP_USER_AGENT
# HTTP_COOKIE
#
# REMOTE_HOST
# REMOTE_ADDR
# REMOTE_USER
# REMOTE_IDENT
#
# GATEWAY_INTERFACE
#
# SERVER_SOFTWARE
# SERVER_NAME
# SERVER_PROTOCOL
# SERVER_PORT
#
# REQUEST_METHOD
#
# PATH_INFO
# PATH_TRANSLATED
#
# SCRIPT_NAME
# QUERY_STRING
#
# AUTH_TYPE
#
# CONTENT_LENGTH
# CONTENT_TYPE
#
# **************************************************************************

(fatal)

(setq code_rx (regcomp "^\(.*"))
(setq comment_rx (regcomp "^;{1,2}"))
(setq include_rx (regcomp "^#[\b\t]*include[\b\t]*\"(([^\"]|\\\\\")+)\""))
(setq suppress_rx (regcomp (join "" "\\\\[" (char 10) (char 13) "]+$")))

(setq current_expr ())

(defun process_expr (line)
   (setq current_expr (append current_expr (list line)))
   (setq line (getline))

   (unless (or (eq line 0) (match comment_rx line))
      (tailcall 0 line))

   (unless (child_running)
      (cgi_warn (date) ": expander.munger: inferior lisp exited prematurely, restarting.")
      (child_open "/usr/local/bin/munger")
      (child_write "(noprinter)" (char 13)))

   (apply child_write current_expr)
   (setq current_expr ())

   (when (eq (length (chomp line)) 2)
      (when (child_wait)
         (do
            (cgi_print (child_read))

            (unless (child_ready)
               (pause 999999))

            (child_ready)))))
   
(defun include (line)
   (let ((m (matches include_rx line)))
      (unless (and m (cadr m))
         (cgi_warn "filter.munger: malformed #include line.")
         (exit 1))

      (with_input_file (cadr m)
         (while (setq m (getline))
            (cgi_print m)))))
   
(child_open "/usr/local/bin/munger")
(child_write "(noprinter)" (char 13))

(letn ((back_rx (regcomp "\\\\"))
       (quot_rx (regcomp "\""))
       (escape (lambda (x) (substitute quot_rx "\\\\\"" (substitute back_rx "\\\\\\\\" x 0) 0)))
       (val 0))

   (defun propagate ()
      (foreach
         (lambda (item)
            (child_write (concat "(setenv \"" (escape item) "\" \""
               (if (setq val (cgi_getenv item)) (escape val) "") "\")")))
      
         '("HTTP_USER_AGENT" 
           "HTTP_COOKIE"

           "REMOTE_HOST" 
           "REMOTE_ADDR" 
           "REMOTE_USER"
           "REMOTE_IDENT"
           
           "GATEWAY_INTERFACE"

           "SERVER_SOFTWARE"
           "SERVER_NAME"
           "SERVER_PROTOCOL"
           "SERVER_PORT"

           "REQUEST_METHOD"

           "PATH_INFO"
           "PATH_TRANSLATED"

           "SCRIPT_NAME"
           "QUERY_STRING"

           "AUTH_TYPE"

           "CONTENT_LENGTH"
           "CONTENT_TYPE"))))
      
(defun main ()
   (cgi_print "Content-Type: text/html; charset=iso-8859-1")
   (cgi_newline)
   (cgi_newline)

   (let ((file (cgi_getenv "PATH_TRANSLATED"))
         (line ""))

      (when file
         (with_input_file file
            (while (setq line (getline))

               (cond ((match code_rx line) (process_expr line))
                     ((match include_rx line) (include line))
                     ((not (match comment_rx line)) 
                      (cgi_print (substitute suppress_rx "" line)))))))))
               
(while (cgi_accept)
   (propagate)
   (main)
   (child_write "(gc)" (char 13))
   (gc))
   
(child_close)
(exit 0)
